/************************************************************************/
/*                                                                      */
/*   knn_distances.c                                                    */
/*                                                                      */
/*   Computes k-nearest neighbors using inner product similarity.       */
/*                                                                      */
/*   Copyright: Thorsten Joachims                                       */
/*   Date: 16.12.97                                                     */
/*                                                                      */
/************************************************************************/

# include <stdio.h>
# include <ctype.h>
# include <signal.h>
# include <math.h>
# include <string.h>
# include <stdlib.h>
# include "./knn.h"

typedef struct dist_dnum {
  long    dnum;	
  double  dist;
} DIST_DNUM;

char resultsdir[200];
char docfile[200];
char modelfile[200];
char distancefile[200];

int cmp_dist();

main (argc, argv)
char *argv[];
{
  WORD **docs,*prior;
  long totdocs=0,max_docs,max_words_doc,i,j;
  long totwords,totdoc,*label,k,temp;
  double dist;
  DIST_DNUM *buffer;
  FILE *predfl;
  KERNEL_PARM kernel_parm;
  MODEL model;

  max_docs=atol(argv[1]);
  max_words_doc=atol(argv[2]);
  totwords=atol(argv[3]);
  k=atol(argv[4]);
  strcpy (resultsdir, argv[5]);
  strcpy (docfile, argv[6]);
  strcpy (modelfile, argv[7]);
  strcpy (distancefile, argv[8]);

  docs = (WORD **)malloc(sizeof(WORD *)*(max_docs+2));
  label = (long *)malloc(sizeof(long)*(max_docs+2));
  buffer = (DIST_DNUM *)malloc(sizeof(DIST_DNUM)*(max_docs+2));
  prior = (WORD *)malloc(sizeof(WORD)*(totwords+2));
  model.supvec = (WORD **)malloc(sizeof(double *)*(max_docs+2));
  model.alpha = (double *)malloc(sizeof(double)*(max_docs+2));

  read_documents(docfile,docs,label,max_words_doc,(long)(max_words_doc*100),&temp,&totdoc);
  read_model(modelfile,&model,totwords,&kernel_parm);
  
  printf("Computing distances.."); fflush(stdout);
  if ((predfl = fopen (distancefile, "w")) == NULL)
  { perror (distancefile); exit (1); }
  for(i=0;i<totdoc;i++) {
    compute_distances(&model,docs[i],k,&kernel_parm,buffer,predfl);
    if(i % 100 == 0) { 
      printf("%ld..",i); fflush(stdout);
    }
  }  
  fclose(predfl);
  printf("done\n");
}

read_model(modelfile,model,totwords,kernel_parm)
char *modelfile;
MODEL *model;
long totwords;
KERNEL_PARM *kernel_parm;
{
  FILE *modelfl;
  long j,i;
  char *line;
  WORD *doc;
  register long wpos;
  long wnum,pos;
  double weight;

  printf("Reading model..."); fflush(stdout);
  doc = (WORD *)malloc(sizeof(WORD)*(totwords+10));
  line = (char *)malloc(sizeof(char)*totwords*40);
  if ((modelfl = fopen (modelfile, "r")) == NULL)
  { perror (modelfile); exit (1); }
  kernel_parm->kernel_type=0;
  kernel_parm->poly_degree=0;
  kernel_parm->rbf_gamma=0;
  kernel_parm->sigmoid_gamma=0;
  kernel_parm->sigmoid_c=0;

  model->sv_num=0;
  fscanf(modelfl,"%[^\n]\n", line);  /* skip prior */
  i=0;
  while(fscanf(modelfl,"%[^\n]\n", line) != EOF) {
    pos=0;
    wpos=0;
    sscanf(line,"%lf",&model->alpha[i]);  /* dummy */
    while(line[++pos]>32);
    while((sscanf(line+pos,"%ld:%lf",&wnum,&weight) != EOF) && (wpos<totwords)) {
      while(line[++pos]>32);
      doc[wpos].wnum=wnum;
      doc[wpos].weight=weight; 
      wpos++;
      /*      fprintf(stdout, "%d:%f ", wnum,weight);  */
    }
    model->supvec[i] = (WORD *)malloc(sizeof(WORD)*(wpos+1));
    for(j=0;j<wpos;j++) (model->supvec[i])[j]=doc[j];
    (model->supvec[i])[wpos].wnum=0;
    model->sv_num++;
    i++;
  }
  fclose(modelfl);
  fprintf(stdout, "OK. (%d training example vectors read)\n", model->sv_num);
}

compute_distances(model,ex,k,kernel_parm,buffer,out)
MODEL *model;
WORD *ex;
long k;
KERNEL_PARM *kernel_parm;
DIST_DNUM *buffer;
FILE *out;
{
  register long i;
  register double dist=0, sum=0;

  for(i=0;i<model->sv_num;i++) { 
    buffer[i].dnum=i;
    buffer[i].dist=kernel(kernel_parm,model->supvec[i],ex);
  }
  qsort(buffer,(size_t)model->sv_num,sizeof(DIST_DNUM),(int (*)(void*,void*))cmp_dist); 
  dist=0;
  sum=0;
  for(i=0;i<k;i++) { 
    fprintf(out,"%.12f:%d ",buffer[i].dist,buffer[i].dnum);
  }
  fprintf(out,"\n");
}

int cmp_dist(a,b)  /* sort in decreasing order */
DIST_DNUM *a,*b;
{
  if(a->dist > b->dist) {
    return(-1);
  }
  else if(a->dist < b->dist) {
    return(1);
  }
  else {
    return(0);
  }
}
